# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2015-2019 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

"""

This module contains reporting constants for Comet

"""

EXPERIMENT_CREATED = "experiment_created"

EXPERIMENT_CREATION_FAILED = "experiment_creation_failed"

FILE_UPLOADED_FAILED = "file_uploaded_failed"

REPO_UPLOADED = "repo_uploaded"

REPO_UPLOADED_FAILED = "repo_uploaded_failed"

ON_EXIT_DIDNT_FINISH_UPLOAD_SDK = "on_exit_didnt_finish_upload_sdk"

WS_ON_OPEN = "ws_on_open"

WS_ON_ERROR = "ws_on_error"

WS_ON_CLOSE = "ws_on_close"
